"""
piddlePS - a PostScript backend for the PIDDLE drawing module

   Magnus Lie Hetland
   
   1999
"""
# some fixups by Chris Lee (cwlee@artsci.wustl.edu)

# $Id: piddlePS.py,v 1.3 2000/02/08 22:59:50 clee Exp $
#
# Ok, PS coords are intially 0 300 translate so that start point is top left of BBox
# thus all y coordinates must be entered in opposite sign to go down

# TODO for version 1.0:
#     X add greyscale image support for smaller images (added to level 2)
#     X add Level2 postscript options
#     X add underline implementation from piddlePDF
#     X ** fix \n breaking in drawString
#     ? Bezier curve thing (Robert Kern's fix is in piddlePDF)
#          Hmm..drawCurve(..) already uses postscript's bezier curveto-- is there anything else
#          to be done? 
#     X drawArc stuff from Eric

# In the Future:
#    _ Base85 ecooding just use hex encoding involves 1:2 expansion of image data vs 4:5
#    _ Don't see a flate/deflate filter for Postscript, jpeg DCTEncode could be added.
#         PIL may have a LZW encoder 
#    _ check Adobe Document struturing conventions
#    _ look at symbol's font metrics--they appear to be a little off
#    X postscript native implementation of drawRoundRect 
#    _ improve underlining placement...doesn't look good for courier and symbol


from piddle import *
import string, cStringIO
import pdfmetrics # for font info

linesep = '\n'

class PSCanvas(Canvas):
    def __init__(self,size=(300,300),name='piddlePS', PostScriptLevel=2):
       Canvas.__init__(self,size,name)
       width, height = self.size = size
       self.filename = name
       if len(name) < 3 or string.lower(name[-3:]) != '.ps':
           self.filename = name + ".ps"

       self.code = []
       
       c = self._currentColor = self.defaultLineColor
       r,g,b = c.red, c.green, c.blue
       
       w = self._currentWidth = self.defaultLineWidth

       self.defaultFont = Font(face='serif') 

       self._currentFont = self.defaultFont
       f = self._findFont(self._currentFont)
       s = self._currentFont.size
       
       self.code.append('''\
%%!PS-Adobe-3.0 EPSF-3.0
%%%%BoundingBox: 0 0 %d %d

%%%% Initialization:

2 setlinecap

0 %d
translate

%s %s %s setrgbcolor
(%s) findfont %s scalefont setfont
%s setlinewidth
''' % (width, height, height, r, g, b, f, s, w))
       # select between postscript level 1 or level 2
       if PostScriptLevel == 1 :
           self.drawImage = self._drawImageLevel1
       elif PostScriptLevel == 2 :
           self.drawImage = self._drawImageLevel2
       else :
           raise 'PostScriptLevelException'
       


    def _findFont(self,font):
        legal = {"helvetica": "Helvetica",
                 "times": "Times",
                 "courier": "Courier",
                 "serif": "Times",
                 "sansserif": "Helvetica",
                 "monospaced": "Courier",
                 "symbol" : "Symbol"} # There are more...
       # symbol related hacks--added to legal list, test for it and don't qualify w/ Roman,
       # bold or italic -cwl 
        from string import lower

        requested = font.face or "Serif" 
        if type(requested) == StringType:
            requested = [requested]

        face = legal["serif"]
        for reqFace in requested:
            if legal.has_key(lower(reqFace)):
                face = legal[lower(reqFace)]
                break

        shapes = {"Helvetica": ["Bold", "Oblique"],
                  "Times": ["Bold", "Italic"],
                  "Courier": ["Bold", "Oblique"],
                  "Symbol": ["", ""] }[face] # ??? what is proper for symbol -cwl

        # print "face = %s" % face
        if face != "Symbol" :
            face = face+"-"

            if font.bold or font.italic:
                if font.bold:
                    face = face+shapes[0]
                if font.italic:
                        face = face+shapes[1]
            else:
                face = face+'Roman'
            
        return face


    def _findExternalFontName(self, font):       #copied from piddlePDF by cwl- hack away!
        """Attempts to return proper font name.
        PDF uses a standard 14 fonts referred to
        by name. Default to self.defaultFont('Helvetica').
        The dictionary allows a layer of indirection to
        support a standard set of PIDDLE font names."""

        piddle_font_map = {
            'Times':'Times',
            'times':'Times',
            'Courier':'Courier',
            'courier':'Courier',
            'helvetica':'Helvetica',
            'Helvetica':'Helvetica',
            'symbol':'Symbol',
            'Symbol':'Symbol',
            'monospaced':'Courier',
            'serif':'Times',
            'sansserif':'Helvetica',
            'ZapfDingbats':'ZapfDingbats',
            'zapfdingbats':'ZapfDingbats',
            'arial':'Helvetica'
            }

        try:
            face = piddle_font_map[string.lower(font.face)]
        except:
            return 'Helvetica'

        name = face + '-'
        if font.bold and face in ['Courier','Helvetica','Times']:
            name = name + 'Bold'
        if font.italic and face in ['Courier', 'Helvetica']:
            name = name + 'Oblique'
        elif font.italic and face == 'Times':
            name = name + 'Italic'

        if name == 'Times-':
            name = name + 'Roman'
        # symbol and ZapfDingbats cannot be modified!

        #trim and return
        if name[-1] == '-':
            name = name[0:-1]
        return name




    def clear(self):
        code.append('showpage')


    def flush(self):
        self.code.append('showpage')
        file = open(self.filename,'w')
        from string import join
        file.write(join(self.code,linesep))
        file.close()
       
    def stringWidth(self, s, font=None):
        "Return the logical width of the string if it were drawn \
        in the current font (defaults to self.font)."
        if not font:
            font = self.defaultFont
        fontname = self._findExternalFontName(font)
        return pdfmetrics.stringwidth(s, fontname) * font.size * 0.001

    
    ### def fontHeight(self, font=None) ### use default piddle method
    # distance between baseline of two lines of text 1.2 * font.size for piddle.py
    # Thus is 1.2 times larger than postscript minimum baseline to baseline separation
    # for single-spaced text.  fontHeight() used internally for drawString() multi-line text

    def fontAscent(self, font=None):
        if not font:
            font = self.defaultFont
        fontname = self._findExternalFontName(font)
        return pdfmetrics.ascent_descent[fontname][0] * 0.001 * font.size


    def fontDescent(self, font=None):
        if not font:
            font = self.defaultFont
        fontname = self._findExternalFontName(font)
        return -pdfmetrics.ascent_descent[fontname][1] * 0.001 * font.size



    def _updateLineColor(self, color):
       color = color or self.defaultLineColor
       if color != self._currentColor:
          self._currentColor = color
          r,g,b = color.red, color.green, color.blue
          self.code.append('%s %s %s setrgbcolor' % (r,g,b))


    def _updateFillColor(self, color):
       color = color or self.defaultFillColor
       if color != self._currentColor:
          self._currentColor = color
          r,g,b = color.red, color.green, color.blue
          self.code.append('%s %s %s setrgbcolor' % (r,g,b))


    def _updateLineWidth(self, width):
       if width == None: width = self.defaultLineWidth
       if width != self._currentWidth:
          self._currentWidth = width
          self.code.append('%s setlinewidth' % width)
          

    def _updateFont(self, font):
       font = font or self.defaultFont
       if font != self._currentFont:
          self._currentFont = font
          f = self._findFont(font)
          s = font.size
          self.code.append('(%s) findfont %s scalefont setfont' % (f, s))


    def drawLine(self, x1, y1, x2, y2, color=None, width=None):
       self._updateLineColor(color)
       self._updateLineWidth(width)
       if self._currentColor != transparent:
          self.code.append('%s %s neg moveto %s %s neg lineto stroke' %
                           (x1, y1, x2, y2))


    def drawLines(self, lineList, color=None, width=None):
       self._updateLineColor(color)
       self._updateLineWidth(width)
       codeline = '%s %s neg moveto %s %s neg lineto stroke'
       if self._currentColor != transparent:
          for line in lineList:
             self.code.append(codeline % line)


    def _escape(self, s):
        # return a copy of string s with special characters in postscript strings
        # escaped" with backslashes."""
        # Have not handled characters that are converted normally in python strings
        # i.e. \n -> newline 
        str = string.replace(s, chr(0x5C), r'\\' )
        str = string.replace(str, '(', '\(' )
        str = string.replace(str, ')', '\)')
        return str 

    # ??? check to see if \n response is handled correctly (should move cursor down)
       
    def _drawStringOneLineNoRot(self, s, x, y, font=None) :
       # PRE: x and y and position at which to place text
       # PRE: helper function, only called from drawString(..)
       text = self._escape(s)
       self.code.append('%s %s neg moveto (%s) show' % (x,y,text))
       if self._currentFont.underline:
           swidth = self.stringWidth(s, self._currentFont)
           ypos =  (0.5 * self.fontDescent(self._currentFont))
           thickness = 0.08 * self._currentFont.size  # relate to font.descent?
           self.code.extend(['%s setlinewidth' % thickness,
                             '0 %s neg rmoveto' % (ypos),
                             '%s 0 rlineto stroke' % -swidth])
       


    def _drawStringOneLine(self, s, x, y, font=None, color=None, angle=0):
       # PRE: assumes that coordinate system has already been set for rotated strings
       # PRE: only meant to be called by drawString(..)
       text = self._escape(s)
       self.code.extend(['%f %f neg moveto (%s) show' % (x, y, text)])

       if self._currentFont.underline:
          swidth = self.stringWidth(s, self._currentFont)
          dy =  (0.5 * self.fontDescent(self._currentFont))
          thickness = 0.08 * self._currentFont.size  # relate to font.descent?
          self.code.extend(['%s setlinewidth' % thickness,
                            '%f %f neg moveto' % (x, dy+y),
                            '%f 0 rlineto stroke' % swidth])


    def drawString(self, s, x, y, font=None, color=None, angle=0):
        self._updateLineColor(color)
        self._updateFont(font)
        if self._currentColor != transparent:

            lines = string.split(s, '\n')   
            lineHeight = self.fontHeight(font)

            if angle == 0 :   # do special case of angle = 0 first. Avoids a bunch of gsave/grestore ops
               for line in lines:
                  self._drawStringOneLineNoRot(line, x, y, font=font)
            else : # general case, rotated text
               self.code.extend([
                  'gsave',
                  '%s %s neg translate' % (x,y),
                  `angle`+' rotate'])
               down = 0
               for line in lines :
                  self._drawStringOneLine(line, 0, 0+down, font, color, angle)
                  down = down + lineHeight
               self.code.extend(['grestore'])

    def drawCurve(self, x1, y1, x2, y2, x3, y3, x4, y4,
                  edgeColor=None, edgeWidth=None, fillColor=None, closed=0):
       codeline = '%s %s neg moveto %s %s neg %s %s neg %s %s neg curveto'
       data = (x1, y1, x2, y2, x3, y3, x4, y4)
       self._updateFillColor(fillColor)
       if self._currentColor != transparent:
          self.code.append((codeline % data) + ' eofill')
       self._updateLineWidth(edgeWidth)
       self._updateLineColor(edgeColor)
       if self._currentColor != transparent:
          self.code.append((codeline % data)
                           + ((closed and ' closepath') or '')
                           + ' stroke')

    ########################################################################################

    def drawRoundRect(self, x1,y1, x2,y2, rx=8, ry=8,
                      edgeColor=None, edgeWidth=None, fillColor=None):
        "Draw a rounded rectangle between x1,y1, and x2,y2, \
        with corners inset as ellipses with x radius rx and y radius ry. \
        These should have x1<x2, y1<y2, rx>0, and ry>0."
        # Path is drawn in counter-clockwise direction"
        
        x1, x2 = min(x1,x2), max(x1, x2) # from piddle.py
        y1, y2 = min(y1,y2), max(y1, y2)

        # Note: arcto command draws a line from current point to beginning of arc
        # save current matrix, translate to center of ellipse, scale by rx ry, and draw
        # a circle of unit radius in counterclockwise dir, return to original matrix
        # arguments are (cx, cy, rx, ry, startAngle, endAngle)
        ellipsePath = 'matrix currentmatrix %s %s neg translate %s %s scale 0 0 1 %s %s arc setmatrix' 

        # choice between newpath and moveto beginning of arc
        # go with newpath for precision, does this violate any assumptions in code???
        # rrcode = ['%s %s neg moveto' % (x1+rx, y1)]  # this also works
        rrcode = ['newpath'] # Round Rect code path
        # upper left corner ellipse is first
        rrcode.append(ellipsePath % (x1+rx, y1+ry, rx, ry, 90, 180))
        rrcode.append(ellipsePath % (x1+rx, y2-ry, rx, ry, 180, 270))
        rrcode.append(ellipsePath % (x2-rx, y2-ry, rx, ry, 270, 360))
        rrcode.append(ellipsePath % (x2-rx, y1+ry, rx, ry, 0,  90) )
        rrcode.append('closepath')
        
        # This is what you are required to do to take care of all color cases
        # should fix this so it doesn't define path twice, just use gsave if need
        # to fill and stroke path-need to figure out this system
        self._updateFillColor(fillColor)
        if self._currentColor != transparent:
            self.code.extend(rrcode)
            self.code.append("eofill")
        self._updateLineWidth(edgeWidth)
        self._updateLineColor(edgeColor)
        if self._currentColor != transparent:
            self.code.extend(rrcode)
            self.code.append("stroke")



        
    def drawEllipse(self, x1,y1, x2,y2, edgeColor=None, edgeWidth=None, 
                    fillColor=None):
        "Draw an orthogonal ellipse inscribed within the rectangle x1,y1,x2,y2. \
        These should have x1<x2 and y1<y2."
        #Just invoke drawArc to actually draw the ellipse
        self.drawArc(x1,y1, x2,y2, edgeColor=edgeColor, edgeWidth=edgeWidth,
                fillColor=fillColor)

    def drawArc(self, x1,y1, x2,y2, startAng=0, extent=360, edgeColor=None,
                edgeWidth=None, fillColor=None):
        "Draw a partial ellipse inscribed within the rectangle x1,y1,x2,y2, \
        starting at startAng degrees and covering extent degrees.   Angles \
        start with 0 to the right (+x) and increase counter-clockwise. \
        These should have x1<x2 and y1<y2."
        #calculate centre of ellipse
        x, y = (x1+x2)/2.0, (y1+y2)/2.0

        codeline = self._genArcCode(x1, y1, x2, y2, startAng, extent) 

        self._updateFillColor(fillColor)
        if self._currentColor != transparent:
            self.code.append('%s %s neg moveto' % (x,y))
            self.code.append(codeline + ' eofill')
        self._updateLineWidth(edgeWidth)
        self._updateLineColor(edgeColor)
        if self._currentColor != transparent:
            self.code.append(codeline + ' stroke')

    def _genArcCode(self, x1, y1, x2, y2, startAng, extent):
        "Calculate the code needed draw an arc"
        #calculate semi-minor and semi-major axes of ellipse
        xScale = abs((x2-x1)/2.0)
        yScale = abs((y2-y1)/2.0)
        #calculate centre of ellipse
        x, y = (x1+x2)/2.0, (y1+y2)/2.0

        codeline = 'matrix currentmatrix '+\
                '%s %s neg translate %s %s scale 0 0 1 %s %s %s '+\
                'setmatrix'

        if extent >= 0:
            arc='arc'
        else:
            arc='arcn'
        data = (x,y, xScale, yScale, startAng, startAng+extent, arc)

        return codeline % data

    
    def drawPolygon(self, pointlist, 
                    edgeColor=None, edgeWidth=None, fillColor=None, closed=0):

       start = pointlist[0]
       pointlist = pointlist[1:]
       
       polyCode = []
       polyCode.append("%s %s neg moveto" % start)
       for point in pointlist:
          polyCode.append("%s %s neg lineto" % point)
       if closed:
          polyCode.append("closepath")

       self._updateFillColor(fillColor)
       if self._currentColor != transparent:
          self.code.extend(polyCode)
          self.code.append("eofill")
       self._updateLineWidth(edgeWidth)
       self._updateLineColor(edgeColor)
       if self._currentColor != transparent:
          self.code.extend(polyCode)
          self.code.append("stroke")


    def drawFigure(self, partList,
                   edgeColor=None, edgeWidth=None, fillColor=None, closed=0):
       
       figureCode = []
       first = 1
       
       for part in partList:
           op = part[0]
           args = list(part[1:])

           if op == figureLine:
               if first:
                   first = 0
                   figureCode.append("%s %s neg moveto" % tuple(args[:2]))
               else:
                   figureCode.append("%s %s neg lineto" % tuple(args[:2]))
               figureCode.append("%s %s neg lineto" % tuple(args[2:]))

           elif op == figureArc:
               first = 0
               x1,y1,x2,y2,startAngle,extent = args[:6]
               figureCode.append(self._genArcCode(x1,y1,x2,y2,startAngle,extent))

           elif op == figureCurve:
               if first:
                   first = 0
                   figureCode.append("%s %s neg moveto" % tuple(args[:2]))
               else:
                   figureCode.append("%s %s neg lineto" % tuple(args[:2]))
               figureCode.append("%s %s neg %s %s neg %s %s neg curveto" % tuple(args[2:]))
           else:
               raise TypeError, "unknown figure operator: "+op

       if closed:
           figureCode.append("closepath")

       self._updateFillColor(fillColor)
       if self._currentColor != transparent:
          self.code.extend(figureCode)
          self.code.append("eofill")
       self._updateLineWidth(edgeWidth)
       self._updateLineColor(edgeColor)
       if self._currentColor != transparent:
          self.code.extend(figureCode)
          self.code.append("stroke")

    ############################################################################################
    # drawImage(self. image, x1, y1, x2=None, y2=None) is now defined by either _drawImageLevel1
    #    ._drawImageLevel2, the choice is made in .__init__ depending on option 


    def _drawImageLevel1(self, image, x1, y1, x2=None,y2=None):
        # Postscript Level1 version available for fallback mode when Level2 doesn't work
       """drawImage(self.image,x1,y1,x2=None,y2=None) : If x2 and y2 are ommitted, they are
       calculated from image size.  (x1,y1) is upper left of image, (x2,y2) is lower right of
       image in piddle coordinates."""
       try:
           import Image
       except ImportError:
           print 'Python Imaging Library not available'
           return
       # For now let's start with 24 bit RGB images (following piddlePDF again)
       print "Trying to drawImage in piddlePS"
       component_depth = 8
       myimage = image.convert('RGB')
       imgwidth, imgheight = myimage.size
       if not x2:
            x2 = imgwidth + x1
       if not y2:
            y2 = y1 + imgheight
       drawwidth = x2 - x1
       drawheight = y2 - y1
       print 'Image size (%d, %d); Draw size (%d, %d)' % (imgwidth, imgheight, drawwidth, drawheight)
       # now I need to tell postscript how big image is

       # "image operators assume that they receive sample data from
       # their data source in x-axis major index order.  The coordinate
       # of the lower-left corner of the first sample is (0,0), of the
       # second (1,0) and so on" -PS2 ref manual p. 215
       #
       # The ImageMatrix maps unit squre of user space to boundary of the source image
       #

       # The CurrentTransformationMatrix (CTM) maps the unit square of
       # user space to the rect...on the page that is to receive the
       # image. A common ImageMatrix is [width 0 0 -height 0 height]
       # (for a left to right, top to bottom image )
       
       # first let's map the user coordinates start at offset x1,y1 on page

       self.code.extend([
           'gsave',
           '%s %s translate' % (x1,-y1 - drawheight), # need to start are lower left of image
           '%s %s scale' % (drawwidth,drawheight),
           '/scanline %d 3 mul string def' % imgwidth  # scanline by multiples of image width
           ])

       # now push the dimensions and depth info onto the stack
       # and push the ImageMatrix to map the source to the target rectangle (see above)
       # finally specify source (PS2 pp. 225 ) and by exmample
       self.code.extend([
           '%s %s %s' % (imgwidth, imgheight, component_depth),
           '[%s %s %s %s %s %s]' % (imgwidth, 0, 0, -imgheight, 0, imgheight),
           '{ currentfile scanline readhexstring pop } false 3',
           'colorimage '
           ])

       # data source output--now we just need to deliver a hex encode
       # series of lines of the right overall size can follow
       # piddlePDF again

       rawimage = myimage.tostring()
       assert(len(rawimage) == imgwidth*imgheight, 'Wrong amount of data for image') 
       #compressed = zlib.compress(rawimage) # no zlib at moment
       hex_encoded = self._AsciiHexEncode(rawimage)
       
       # write in blocks of 78 chars per line
       outstream = cStringIO.StringIO(hex_encoded)

       dataline = outstream.read(78)
       while dataline <> "":
           self.code.append(dataline)
           dataline= outstream.read(78)
       self.code.append('% end of image data') # for clarity
       self.code.append('grestore') # return coordinates to normal

    # end of drawImage

       
    def _AsciiHexEncode(self, input):  # also based on piddlePDF
        "Helper function used by images"
        output = cStringIO.StringIO()
        for char in input:
            output.write('%02x' % ord(char))
        output.reset()
        return output.read()

    def _drawImageLevel2(self, image, x1,y1, x2=None,y2=None): # Postscript Level2 version
        try:
            import Image
        except ImportError:
            print 'Python Imaging Library not available'
            return
               # I don't have zlib -cwl
#         try:
#             import zlib
#         except ImportError:
#             print 'zlib not available'
#             return
       

        ### what sort of image are we to draw
        if image.mode=='L' :
            print 'found image.mode= L'
            imBitsPerComponent = 8
            imNumComponents = 1
            myimage = image 
        elif image.mode == '1':
            print 'found image.mode= 1'
            myimage = image.convert('L')
            imNumComponents = 1
            myimage = image 
        else :
            myimage = image.convert('RGB')
            imNumComponents = 3
            imBitsPerComponent = 8

        imwidth, imheight = myimage.size
        # print 'imwidth = %s, imheight = %s' % myimage.size
        if not x2:
            x2 = imwidth + x1
        if not y2:
            y2 = y1 + imheight
        drawwidth = x2 - x1
        drawheight = y2 - y1
        self.code.extend([
            'gsave',
            '%s %s translate' % (x1,-y1 - drawheight), # need to start are lower left of image
            '%s %s scale' % (drawwidth,drawheight)])

        if imNumComponents == 3 :
            self.code.append('/DeviceRGB setcolorspace')
        elif imNumComponents == 1 :
            self.code.append('/DeviceGray setcolorspace')
            print 'setting colorspace gray'
        # create the image dictionary
        self.code.append("""
<<
/ImageType 1
/Width %d /Height %d  %% dimensions of source image
/BitsPerComponent %d""" % (imwidth, imheight, imBitsPerComponent) )

        if imNumComponents == 1:
            self.code.append('/Decode [0 1]')
        if imNumComponents == 3:
            self.code.append('/Decode [0 1 0 1 0 1]  %% decode color values normally')

        self.code.extend(['/ImageMatrix [%s 0 0 %s 0 %s]' % (imwidth, -imheight, imheight),
                          '/DataSource currentfile /ASCIIHexDecode filter',
                          '>> % End image dictionary',
                          'image'])
        # after image operator just need to dump image dat to file as hexstring
        rawimage = myimage.tostring()
        assert(len(rawimage) == imwidth*imheight, 'Wrong amount of data for image') 
        #compressed = zlib.compress(rawimage) # no zlib at moment
        hex_encoded = self._AsciiHexEncode(rawimage)
       
        # write in blocks of 78 chars per line
        outstream = cStringIO.StringIO(hex_encoded)

        dataline = outstream.read(78)
        while dataline <> "":
            self.code.append(dataline)
            dataline= outstream.read(78)
        self.code.append('> % end of image data') # > is EOD for hex encoded filterfor clarity
        self.code.append('grestore') # return coordinates to normal

        
        
                         
